"""
xmppwb._version
~~~~~~~~~~~~~~~

This module contains the xmppwb version.

:copyright: (c) 2016 by saqura.
:license: MIT, see LICENSE for more details.
"""

# The version is in a seperate file so that we can easily extract it from
# setup.py without importing the whole package.
# See: http://python-packaging-user-guide.readthedocs.org/en/latest/single_source_version/

__version__ = '0.4.0'
